@@ -6,13 +6,16 @@ import os
 
 from typing import Optional, TypeVar, Union, cast
 
-from .jdconstants import *
+from .constants import *
+from .system.constants import *
 from .events import *
 from .packet import *
 from .transport import Transport
 
 import jacdac.util as util
-from .util import now, log, logv
+from .util import now, log, logv, unpack
+from .control.constants import *
+from .pack import PackType, jdpack, jdunpack
 
 
 EV_CHANGE = "change"
@@ -48,15 +51,18 @@ def _service_matches(dev: 'Device', serv: bytearray):
 
 
 class Bus(EventEmitter):
-    def __init__(self, transport: Transport, devid: str = None) -> None:
+    """A Jacdac bus that managed devices, service client, registers...
+    """
+
+    def __init__(self, transport: Transport, *, device_id: str = None) -> None:
         super().__init__(self)
         self.devices: list['Device'] = []
         self.unattached_clients: list['Client'] = []
         self.all_clients: list['Client'] = []
         self.servers: list['Server'] = []
-        if devid is None:
-            devid = random.randbytes(8).hex()
-        self.self_device = Device(self, devid, bytearray(4))
+        if device_id is None:
+            device_id = random.randbytes(8).hex()
+        self.self_device = Device(self, device_id, bytearray(4))
         self.process_thread = threading.Thread(target=self._process_task)
         self.transport = transport
         self._sendq: queue.Queue[bytes] = queue.Queue()
@@ -64,10 +70,10 @@ class Bus(EventEmitter):
 
         self.loop = asyncio.new_event_loop()
 
-        def handler(loop, context): # type: ignore
-            self.loop.default_exception_handler(context) # type: ignore
+        def handler(loop, context):  # type: ignore
+            self.loop.default_exception_handler(context)  # type: ignore
             os._exit(10)
-        self.loop.set_exception_handler(handler) # type: ignore
+        self.loop.set_exception_handler(handler)  # type: ignore
 
         self.sender_thread = threading.Thread(target=self._sender)
         self.sender_thread.start()
@@ -78,6 +84,12 @@ class Bus(EventEmitter):
 
         log("starting bus, self={}", self.self_device)
 
+    def run(self, cb: Callable[..., None], *args: Any):
+        if self.process_thread is threading.current_thread():
+            cb(*args)
+        else:
+            self.loop.call_soon(cb, *args)
+
     def _sender(self):
         while True:
             pkt = self._sendq.get()
@@ -87,12 +99,17 @@ class Bus(EventEmitter):
         loop = self.loop
         asyncio.set_event_loop(loop)
 
-        from . import ctrl
-        ctrls = ctrl.CtrlServer(self)  # attach control server
+        # TODO: what's the best way to import these things
+        from .control.server import ControlServer
+        ctrls = ControlServer(self)  # attach control server
+
+        # TODO: make this optional.
+        from .unique_brain.server import UniqueBrainServer
+        brain = UniqueBrainServer(self)
 
         def keep_task(t: asyncio.Task[None]):
             if t.done():
-                t.result() # throw exception if needed
+                t.result()  # throw exception if needed
                 return False
             return True
 
@@ -100,13 +117,11 @@ class Bus(EventEmitter):
             self.emit(EV_SELF_ANNOUNCE)
             self._gc_devices()
             ctrls.queue_announce()
-            self.pending_tasks = [x for x in self.pending_tasks if keep_task(x)]
+            self.pending_tasks = [
+                x for x in self.pending_tasks if keep_task(x)]
             loop.call_later(0.500, announce)
         loop.call_later(0.500, announce)
 
-        from . import sample
-        sample.acc_sample(self)
-
         def process_later(pkt: bytes):
             loop.call_soon_threadsafe(self.process_frame, pkt)
         self.transport.on_receive = process_later
@@ -303,18 +318,46 @@ class Bus(EventEmitter):
 
 
 class RawRegisterClient(EventEmitter):
-    def __init__(self, client: 'Client', code: int) -> None:
+    """A Jacdac register client
+    """
+
+    def __init__(self, client: 'Client', code: int, pack_format: Union[str, None]) -> None:
         super().__init__(client.bus)
         self.code = code
         self._data: Optional[bytearray] = None
         self._refreshed_at = 0
         self.client = client
+        self.pack_format = pack_format
 
     def current(self, refresh_ms: int = 500):
         if self._refreshed_at + refresh_ms >= now():
             return self._data
         return None
 
+    def unpacked(self) -> tuple[Any, ...]:
+        data = self.query_no_wait()
+        if data and self.pack_format:
+            return unpack(data, self.pack_format)
+        return ()
+
+    def value(self, index: int):
+        values = self.unpacked()
+        if (len(values) > index):
+            return values[index]
+        else:
+            return None
+
+    def set_value(self, index: int, value: Any):
+        # TODO:
+        pass
+
+    def float_value(self, index: int = 0, scale: int = 1) -> Union[float, None]:
+        value = self.value(index)
+        if value is None:
+            return None
+        else:
+            return float(value) * scale
+
     def _query(self):
         pkt = JDPacket(cmd=JD_GET(self.code))
         self.client.send_cmd(pkt)
@@ -322,28 +365,32 @@ class RawRegisterClient(EventEmitter):
     def refresh(self):
         if self._refreshed_at < 0:
             return  # already in progress
-        prev_data = self._data
-        self._refreshed_at = -1
 
-        def final_check():
-            if prev_data is self._data:
-                # if we still didn't get any data, emit "change" event, so that queries can time out
-                self._data = None
-                self._refreshed_at = 0
-                self.emit(EV_CHANGE)
+        def do_refresh():
+            prev_data = self._data
+            self._refreshed_at = -1
 
-        def second_refresh():
-            if prev_data is self._data:
-                self._query()
-                self.bus.loop.call_later(0.100, final_check)
+            def final_check():
+                if prev_data is self._data:
+                    # if we still didn't get any data, emit "change" event, so that queries can time out
+                    self._data = None
+                    self._refreshed_at = 0
+                    self.emit(EV_CHANGE)
 
-        def first_refresh():
-            if prev_data is self._data:
-                self._query()
-                self.bus.loop.call_later(0.050, second_refresh)
+            def second_refresh():
+                if prev_data is self._data:
+                    self._query()
+                    self.bus.loop.call_later(0.100, final_check)
 
-        self._query()
-        self.bus.loop.call_later(0.020, first_refresh)
+            def first_refresh():
+                if prev_data is self._data:
+                    self._query()
+                    self.bus.loop.call_later(0.050, second_refresh)
+
+            self._query()
+            self.bus.loop.call_later(0.020, first_refresh)
+
+        self.bus.run(do_refresh)
 
     # can't be called from event handlers!
     def query(self, refresh_ms: int = 500):
@@ -354,7 +401,7 @@ class RawRegisterClient(EventEmitter):
         self.wait_for(EV_CHANGE)
         if self._data is None:
             raise RuntimeError(
-                "Can't read reg #{} (from {})".format(self.code, self.client))
+                "Can't read reg #{} (from {})".format(hex(self.code), self.client))
         return self._data
 
     async def query_async(self, refresh_ms: int = 500):
@@ -365,7 +412,7 @@ class RawRegisterClient(EventEmitter):
         await self.event(EV_CHANGE)
         if self._data is None:
             raise RuntimeError(
-                "Can't read reg #{} (from {})".format(self.code, self.client))
+                "Can't read reg #{} (from {})".format(hex(self.code), self.client))
         return self._data
 
     def query_no_wait(self, refresh_ms: int = 500):
@@ -471,10 +518,11 @@ class Server(EventEmitter):
 
 
 class Client(EventEmitter):
-    def __init__(self, bus: Bus, service_class: int, role: str) -> None:
+    def __init__(self, bus: Bus, service_class: int, pack_formats: dict[int, str], role: str) -> None:
         super().__init__(bus)
         self.broadcast = False
         self.service_class = service_class
+        self.pack_formats = pack_formats
         self.service_index = None
         self.device: Optional['Device'] = None
         self.current_device:  Optional['Device'] = None
@@ -497,7 +545,9 @@ class Client(EventEmitter):
     def register(self, code: int):
         r = self._lookup_register(code)
         if r is None:
-            r = RawRegisterClient(self, code)
+            pack_format = self.pack_formats[code]
+            # TODO: error policy?
+            r = RawRegisterClient(self, code, pack_format)
             self._registers.append(r)
         return r
 
@@ -521,8 +571,16 @@ class Client(EventEmitter):
         pkt._header[3] |= JD_FRAME_FLAG_COMMAND
         self.bus._send_core(pkt)
 
-    def on_attach(self):
-        pass
+    def send_cmd_packed(self, cmd: int, args: PackType = None):
+        if args is None:
+            pkt = JDPacket(cmd=cmd)
+        else:
+            if not cmd in self.pack_formats:
+                raise RuntimeError("unknown data format")
+            fmt = self.pack_formats[cmd]
+            data = jdpack(fmt, args)
+            pkt = JDPacket(cmd=cmd, data=data)
+        self.send_cmd(pkt)
 
     def _attach(self, dev: 'Device', service_idx: int):
         assert self.device is None
@@ -547,6 +605,30 @@ class Client(EventEmitter):
             self.bus.clear_attach_cache()
         self.emit(EV_DISCONNECTED)
 
+    def on_event(self, code: int, handler: EventHandlerFn) -> UnsubscribeFn:
+        """Registers an event handler for the given event code
+
+        Args:
+            code (int): event identifier code
+            handler (EventHandlerFn): function to run with decoded event data and packet
+
+        Returns:
+            UnsubscribeFn: function to call to unregister handler
+        """
+        if code in self.pack_formats:
+            fmt = self.pack_formats[code]
+        else:
+            fmt = None
+
+        def cb(pkt: JDPacket) -> None:
+            if pkt.event_code == code:
+                if fmt is None:
+                    data = []
+                else:
+                    data = jdunpack(pkt.data, fmt)
+                handler(data)
+        return self.on(EV_EVENT, cb)
+
 
 _JD_CONTROL_ANNOUNCE_FLAGS_RESTART_COUNTER_STEADY = const(0xf)
 
@@ -565,7 +647,8 @@ class Device(EventEmitter):
     @property
     def ctrl_client(self):
         if self._ctrl_client is None:
-            self._ctrl_client = Client(self.bus, 0, "")
+            self._ctrl_client = Client(
+                self.bus, JD_SERVICE_CLASS_CONTROL, JD_CONTROL_PACK_FORMATS, "")
             self._ctrl_client._attach(self, 0)
         return self._ctrl_client
 
@@ -611,7 +694,7 @@ class Device(EventEmitter):
         if not role or role == self.device_id or role == "{}:{}".format(self.device_id, service_idx):
             return True
         return True
-        # return jacdac._rolemgr.getRole(self.deviceId, serviceIdx) == role
+        # TODO: return jacdac._rolemgr.getRole(self.deviceId, serviceIdx) == role
 
     @property
     def num_service_classes(self):
